import numpy as np
import psyneulink as pnl
import pytest

from queue import Queue


class TestRPC:

    def test_transfer_mech(self):

        T_1 = pnl.TransferMechanism(name='log_test_T_1', input_shapes=2)
        T_2 = pnl.TransferMechanism(name='log_test_T_2', input_shapes=2)
        PS = pnl.Composition(name='log_test_PS', pathways=[T_1, T_2])
        con_with_rpc_pipeline = pnl.Context(rpc_pipeline=Queue(), execution_id=PS)

        T_1.set_log_conditions('mod_noise')
        T_1.set_log_conditions(pnl.RESULT)

        T_1.set_delivery_conditions('mod_noise')
        T_1.set_delivery_conditions(pnl.RESULT)

        PS.run(inputs={T_1: [0, 0]}, context=con_with_rpc_pipeline)
        PS.run(inputs={T_1: [1, 2]}, context=con_with_rpc_pipeline)
        PS.run(inputs={T_1: [3, 4]}, context=con_with_rpc_pipeline)

        # assert T_1.log.print_entries() ==
        # test_log.py::TestLog::test_log
        # Log for log_test_T_1:
        # Logged Item:   Time       Context                    Value
        # 'RESULT'       0:0:0:0   'PROCESSING, COMPOSI...   [0. 0.]
        # 'RESULT'       1:0:0:0   'PROCESSING, COMPOSI...   [1. 2.]
        # 'RESULT'       2:0:0:0   'PROCESSING, COMPOSI...   [3. 4.]
        # 'mod_noise'    0:0:0:0   'PROCESSING, COMPOSI...   [0.]
        # 'mod_noise'    1:0:0:0   'PROCESSING, COMPOSI...   [0.]
        # 'mod_noise'    2:0:0:0   'PROCESSING, COMPOSI...   [0.]

        expected = [
            [[0], [1], [2]],
            [[0.], [0.], [0.]],
            [[0.], [0.], [0.]],
            [[0.], [0.], [0.]],
            [[0.0], [0.0], [0.0]],
            [[0., 0.], [1., 2.], [3., 4.]],
        ]

        actual = []
        pipeline = con_with_rpc_pipeline.rpc_pipeline
        while not pipeline.empty():
            actual.append(pipeline.get())
        t_1_entries = [i for i in actual if i.componentName == 'log_test_T_1']
        noise = [i for i in t_1_entries if i.parameterName == 'noise']
        results = [i for i in t_1_entries if i.parameterName == 'RESULT']
        assert all([
            noise[0].time == '0:0:0:0', noise[0].value.data == [0],results[0].value.data == [0.0, 0.0],
            noise[1].time == '1:0:0:0', noise[1].value.data == [0],results[1].value.data == [1.0, 2.0],
            noise[2].time == '2:0:0:0', noise[2].value.data == [0],results[2].value.data == [3.0, 4.0],
        ])

    def test_delivery_initialization(self):
        T = pnl.TransferMechanism(
                prefs={pnl.DELIVERY_PREF: pnl.PreferenceEntry(pnl.LogCondition.EXECUTION, pnl.PreferenceLevel.INSTANCE)}
        )
        comp = pnl.Composition(name='comp', nodes=[T])
        con_with_rpc_pipeline = pnl.Context(rpc_pipeline=Queue(), execution_id=comp)
        pipeline = con_with_rpc_pipeline.rpc_pipeline
        comp.run([1], context=con_with_rpc_pipeline)
        actual = []
        while not pipeline.empty():
            actual.append(pipeline.get())
        assert all([
            len(actual) == 1,
            actual[0].time == '0:0:0:0',
            actual[0].value.shape == [1, 1],
            actual[0].value.data == [1.0]
        ])

    def test_run_resets(self):
        T1 = pnl.TransferMechanism(name='log_test_T1',
                                   input_shapes=2)
        T2 = pnl.TransferMechanism(name='log_test_T2',
                                   input_shapes=2)
        COMP = pnl.Composition(name='COMP', pathways=[T1, T2])
        con_with_rpc_pipeline = pnl.Context(rpc_pipeline=Queue(), execution_id=COMP)
        pipeline = con_with_rpc_pipeline.rpc_pipeline
        T1.set_delivery_conditions('mod_slope')
        T2.set_delivery_conditions('value')
        COMP.run(inputs={T1: [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]},
                 context=con_with_rpc_pipeline)
        pipeline = con_with_rpc_pipeline.rpc_pipeline
        actual = []
        while not pipeline.empty():
            actual.append(pipeline.get())
        assert all([i.context == 'COMP' for i in actual])
        np.testing.assert_allclose([
            np.ndarray(shape=np.array(actual[1].value.shape), buffer=np.array(actual[1].value.data)),
            np.ndarray(shape=np.array(actual[3].value.shape), buffer=np.array(actual[3].value.data)),
            np.ndarray(shape=np.array(actual[5].value.shape), buffer=np.array(actual[5].value.data)),
        ], [[[1.0, 2.0]], [[3.0, 4.0]], [[5.0, 6.0]]])

        COMP.run(inputs={T1: [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]},
                 context=con_with_rpc_pipeline)
        actual = []
        while not pipeline.empty():
            actual.append(pipeline.get())
        np.testing.assert_allclose([
            np.ndarray(shape=np.array(actual[1].value.shape), buffer=np.array(actual[1].value.data)),
            np.ndarray(shape=np.array(actual[3].value.shape), buffer=np.array(actual[3].value.data)),
            np.ndarray(shape=np.array(actual[5].value.shape), buffer=np.array(actual[5].value.data)),
        ], [[[1.0, 2.0]], [[3.0, 4.0]], [[5.0, 6.0]]])

    def test_log_dictionary_with_time(self):
        T1 = pnl.TransferMechanism(name='log_test_T1',
                                   input_shapes=2)
        T2 = pnl.TransferMechanism(name='log_test_T2',
                                   function=pnl.Linear(slope=2.0),
                                   input_shapes=2)
        COMP = pnl.Composition(name='log_test_COMP', pathways=[T1, T2])
        con_with_rpc_pipeline = pnl.Context(rpc_pipeline=Queue(), execution_id=COMP)
        pipeline = con_with_rpc_pipeline.rpc_pipeline

        T1.set_delivery_conditions('mod_slope')
        T1.set_delivery_conditions(pnl.RESULT)
        T1.set_delivery_conditions(pnl.VALUE)

        T2.set_delivery_conditions('mod_slope')
        T2.set_delivery_conditions(pnl.RESULT)
        T2.set_delivery_conditions(pnl.VALUE)

        # RUN ZERO  |  TRIALS ZERO, ONE, TWO ----------------------------------

        COMP.run(inputs={T1: [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]},
                 context=con_with_rpc_pipeline)

        actual = []
        while not pipeline.empty():
            actual.append(pipeline.get())

        t1_slope_entries = [i for i in actual if i.parameterName == pnl.SLOPE and i.componentName == 'log_test_T1']
        t1_slope_values = [np.ndarray(shape=np.array(i.value.shape), buffer=np.array(i.value.data)) for i in t1_slope_entries]
        t1_slope_times = [i.time for i in t1_slope_entries]

        t2_slope_entries = [i for i in actual if i.parameterName == pnl.SLOPE and i.componentName == 'log_test_T2']
        t2_slope_values = [np.ndarray(shape=np.array(i.value.shape), buffer=np.array(i.value.data)) for i in t2_slope_entries]
        t2_slope_times = [i.time for i in t2_slope_entries]

        t1_result_entries = [i for i in actual if i.parameterName == pnl.RESULT and i.componentName == 'log_test_T1']
        t1_result_values = [np.ndarray(shape=np.array(i.value.shape), buffer=np.array(i.value.data)) for i in t1_result_entries]
        t1_result_times = [i.time for i in t1_result_entries]

        t2_result_entries = [i for i in actual if i.parameterName == pnl.RESULT and i.componentName == 'log_test_T2']
        t2_result_values = [np.ndarray(shape=np.array(i.value.shape), buffer=np.array(i.value.data)) for i in t2_result_entries]
        t2_result_times = [i.time for i in t2_result_entries]

        t1_value_entries = [i for i in actual if i.parameterName == pnl.VALUE and i.componentName == 'log_test_T1']
        t1_value_values = [np.ndarray(shape=np.array(i.value.shape), buffer=np.array(i.value.data)) for i in t1_value_entries]
        t1_value_times = [i.time for i in t1_value_entries]

        t2_value_entries = [i for i in actual if i.parameterName == pnl.VALUE and i.componentName == 'log_test_T2']
        t2_value_values = [np.ndarray(shape=np.array(i.value.shape), buffer=np.array(i.value.data)) for i in t2_value_entries]
        t2_value_times = [i.time for i in t2_value_entries]

        # Test execution contexts for all entries

        assert all([True if i.context == COMP.default_execution_id else False for i in actual])

        # T1 log after zero-th run -------------------------------------------

        expected_times_T1 = ['0:0:0:0', '0:1:0:0', '0:2:0:0']
        expected_values_T1 = [[[1.0, 2.0]], [[3.0, 4.0]], [[5.0, 6.0]]]
        expected_slopes_T1 = [[1.0], [1.0], [1.0]]
        expected_results_T1 = [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]

        assert expected_times_T1 == t1_result_times == t1_slope_times == t1_value_times
        np.testing.assert_allclose(expected_values_T1, t1_value_values)
        np.testing.assert_allclose(expected_results_T1, t1_result_values)
        np.testing.assert_allclose(expected_slopes_T1, t1_slope_values)

        # T2 log after zero-th run --------------------------------------------

        expected_times_T2 = ['0:0:0:1', '0:1:0:1', '0:2:0:1']
        expected_values_T2 = [[[2.0, 4.0]], [[6.0, 8.0]], [[10.0, 12.0]]]
        expected_slopes_T2 = [[2.0], [2.0], [2.0]]
        expected_results_T2 = [[2.0, 4.0], [6.0, 8.0], [10.0, 12.0]]

        assert expected_times_T2 == t2_result_times == t2_slope_times == t2_value_times
        np.testing.assert_allclose(expected_values_T2, t2_value_values)
        np.testing.assert_allclose(expected_results_T2, t2_result_values)
        np.testing.assert_allclose(expected_slopes_T2, t2_slope_values)

        # RUN ONE  |  TRIALS ZERO, ONE, TWO -------------------------------------

        COMP.run(inputs={T1: [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]},
                 context=con_with_rpc_pipeline)

        actual = []
        while not pipeline.empty():
            actual.append(pipeline.get())

        t1_slope_entries = [i for i in actual if i.parameterName == pnl.SLOPE and i.componentName == 'log_test_T1']
        t1_slope_values = [np.ndarray(shape=np.array(i.value.shape), buffer=np.array(i.value.data)) for i in t1_slope_entries]
        t1_slope_times = [i.time for i in t1_slope_entries]

        t2_slope_entries = [i for i in actual if i.parameterName == pnl.SLOPE and i.componentName == 'log_test_T2']
        t2_slope_values = [np.ndarray(shape=np.array(i.value.shape), buffer=np.array(i.value.data)) for i in t2_slope_entries]
        t2_slope_times = [i.time for i in t2_slope_entries]

        t1_result_entries = [i for i in actual if i.parameterName == pnl.RESULT and i.componentName == 'log_test_T1']
        t1_result_values = [np.ndarray(shape=np.array(i.value.shape), buffer=np.array(i.value.data)) for i in t1_result_entries]
        t1_result_times = [i.time for i in t1_result_entries]

        t2_result_entries = [i for i in actual if i.parameterName == pnl.RESULT and i.componentName == 'log_test_T2']
        t2_result_values = [np.ndarray(shape=np.array(i.value.shape), buffer=np.array(i.value.data)) for i in t2_result_entries]
        t2_result_times = [i.time for i in t2_result_entries]

        t1_value_entries = [i for i in actual if i.parameterName == pnl.VALUE and i.componentName == 'log_test_T1']
        t1_value_values = [np.ndarray(shape=np.array(i.value.shape), buffer=np.array(i.value.data)) for i in t1_value_entries]
        t1_value_times = [i.time for i in t1_value_entries]

        t2_value_entries = [i for i in actual if i.parameterName == pnl.VALUE and i.componentName == 'log_test_T2']
        t2_value_values = [np.ndarray(shape=np.array(i.value.shape), buffer=np.array(i.value.data)) for i in t2_value_entries]
        t2_value_times = [i.time for i in t2_value_entries]

        # T1 log after first run -------------------------------------------

        # Test execution contexts for all entries

        assert all([True if i.context == COMP.default_execution_id else False for i in actual])

        expected_times_T1 = ['1:0:0:0', '1:1:0:0', '1:2:0:0']
        expected_values_T1 = [[[1.0, 2.0]], [[3.0, 4.0]], [[5.0, 6.0]]]
        expected_slopes_T1 = [[1.0], [1.0], [1.0]]
        expected_results_T1 = [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]

        assert expected_times_T1 == t1_result_times == t1_slope_times == t1_value_times
        np.testing.assert_allclose(expected_values_T1, t1_value_values)
        np.testing.assert_allclose(expected_results_T1, t1_result_values)
        np.testing.assert_allclose(expected_slopes_T1, t1_slope_values)

        # T2 log after first run -------------------------------------------

        expected_times_T2 = ['1:0:0:1', '1:1:0:1', '1:2:0:1']
        expected_values_T2 = [[[2.0, 4.0]], [[6.0, 8.0]], [[10.0, 12.0]]]
        expected_slopes_T2 = [[2.0], [2.0], [2.0]]
        expected_results_T2 = [[2.0, 4.0], [6.0, 8.0], [10.0, 12.0]]

        assert expected_times_T2 == t2_result_times == t2_slope_times == t2_value_times
        np.testing.assert_allclose(expected_values_T2, t2_value_values)
        np.testing.assert_allclose(expected_results_T2, t2_result_values)
        np.testing.assert_allclose(expected_slopes_T2, t2_slope_values)

    def test_log_dictionary_with_scheduler(self):
        T1 = pnl.TransferMechanism(name='log_test_T1',
                                   integrator_mode=True,
                                   integration_rate=0.5)
        T2 = pnl.TransferMechanism(name='log_test_T2',
                                   function=pnl.Linear(slope=6.0))
        COMP = pnl.Composition(name='log_test_COMP', pathways=[T1, T2])
        con_with_rpc_pipeline = pnl.Context(rpc_pipeline=Queue(), execution_id=COMP)
        pipeline = con_with_rpc_pipeline.rpc_pipeline

        terminate_trial = {
            pnl.TimeScale.TRIAL: pnl.Threshold(T2, 'value', 5.0, '>=', (0, 0))
        }

        T1.set_delivery_conditions(pnl.VALUE)
        T1.set_delivery_conditions('mod_slope')
        T1.set_delivery_conditions(pnl.RESULT)
        T2.set_delivery_conditions(pnl.VALUE)
        T2.set_delivery_conditions('mod_slope')

        COMP.run(inputs={T1: [[1.0]]}, termination_processing=terminate_trial,
                 context=con_with_rpc_pipeline)

        actual = []
        while not pipeline.empty():
            actual.append(pipeline.get())
        assert all([True if i.context == COMP.default_execution_id else False for i in actual])

        t1_slope_entries = [i for i in actual if i.parameterName == pnl.SLOPE and i.componentName == 'log_test_T1']
        t1_slope_values = [np.ndarray(shape=np.array(i.value.shape), buffer=np.array(i.value.data)) for i in t1_slope_entries]
        t1_slope_times = [i.time for i in t1_slope_entries]

        t2_slope_entries = [i for i in actual if i.parameterName == pnl.SLOPE and i.componentName == 'log_test_T2']
        t2_slope_values = [np.ndarray(shape=np.array(i.value.shape), buffer=np.array(i.value.data)) for i in t2_slope_entries]
        t2_slope_times = [i.time for i in t2_slope_entries]

        t1_result_entries = [i for i in actual if i.parameterName == pnl.RESULT and i.componentName == 'log_test_T1']
        t1_result_values = [np.ndarray(shape=np.array(i.value.shape), buffer=np.array(i.value.data)) for i in t1_result_entries]
        t1_result_times = [i.time for i in t1_result_entries]

        t1_value_entries = [i for i in actual if i.parameterName == pnl.VALUE and i.componentName == 'log_test_T1']
        t1_value_values = [np.ndarray(shape=np.array(i.value.shape), buffer=np.array(i.value.data)) for i in t1_value_entries]
        t1_value_times = [i.time for i in t1_value_entries]

        t2_value_entries = [i for i in actual if i.parameterName == pnl.VALUE and i.componentName == 'log_test_T2']
        t2_value_values = [np.ndarray(shape=np.array(i.value.shape), buffer=np.array(i.value.data)) for i in t2_value_entries]
        t2_value_times = [i.time for i in t2_value_entries]

        # Check values T1

        expected_times_T1 = ['0:0:0:0', '0:0:1:0', '0:0:2:0']
        expected_results_T1 = [[0.5], [0.75], [0.875]]
        expected_values_T1 = [[[0.5]], [[0.75]], [[0.875]]]
        expected_slopes_T1 = [[1], [1], [1]]
        assert expected_times_T1 == t1_result_times == t1_slope_times == t1_value_times
        np.testing.assert_allclose(expected_values_T1, t1_value_values)
        np.testing.assert_allclose(expected_results_T1, t1_result_values)
        np.testing.assert_allclose(expected_slopes_T1, t1_slope_values)

        # Check values T2

        expected_times_T2 = ['0:0:0:1', '0:0:1:1', '0:0:2:1']
        expected_values_T2 = [[[3]], [[4.5]], [[5.25]]]
        expected_slopes_T2 = [[6], [6], [6]]
        assert expected_times_T2 == t2_slope_times == t2_value_times
        np.testing.assert_allclose(expected_values_T2, t2_value_values)
        np.testing.assert_allclose(expected_slopes_T2, t2_slope_values)

    def test_log_dictionary_with_scheduler_many_time_step_increments(self):
        T1 = pnl.TransferMechanism(name='log_test_T1',
                                   integrator_mode=True,
                                   integration_rate=0.05)
        COMP = pnl.Composition(name='log_test_COMP', pathways=[T1])
        con_with_rpc_pipeline = pnl.Context(rpc_pipeline=Queue(), execution_id=COMP)
        pipeline = con_with_rpc_pipeline.rpc_pipeline

        terminate_trial = {
            pnl.TimeScale.TRIAL: pnl.Threshold(T1, 'value', 0.95, '>=')
        }

        T1.set_delivery_conditions(pnl.VALUE)

        COMP.run(inputs={T1: [[1.0]]}, termination_processing=terminate_trial, context=con_with_rpc_pipeline)

        actual = []
        while not pipeline.empty():
            actual.append(pipeline.get())
        assert all([True if i.context == COMP.default_execution_id else False for i in actual])

        t1_value_entries = [i for i in actual if i.parameterName == pnl.VALUE and i.componentName == 'log_test_T1']
        t1_value_values = [np.ndarray(shape=np.array(i.value.shape), buffer=np.array(i.value.data)) for i in
                           t1_value_entries]

        # Check values T1
        assert len(actual) == 59
        assert actual[30].time == '0:0:30:0'
        assert t1_value_values[58] >= 0.95
        assert t1_value_values[57] < 0.95

    def test_log_csv_multiple_contexts(self):
        pipeline = Queue()
        con_X = pnl.Context(execution_id='comp X', rpc_pipeline=pipeline)
        con_Y = pnl.Context(execution_id='comp Y', rpc_pipeline=pipeline)

        A = pnl.TransferMechanism(name='A')
        B = pnl.TransferMechanism(name='B')
        C = pnl.TransferMechanism(name='C')

        C.set_delivery_conditions(pnl.VALUE)

        X = pnl.Composition(name='comp X')
        Y = pnl.Composition(name='comp Y')

        X.add_linear_processing_pathway([A, C])
        Y.add_linear_processing_pathway([B, C])

        # running with manual contexts for consistent output
        # because output is sorted by context
        X.run(inputs={A: 1}, context=con_X)
        Y.run(inputs={B: 2}, context=con_Y)

        actual = []
        while not pipeline.empty():
            actual.append(pipeline.get())

        assert actual[0].context == 'comp X'
        assert actual[0].time == '0:0:0:1'
        assert actual[0].value.data == [1]
        assert actual[1].context == 'comp Y'
        assert actual[1].time == '0:0:0:1'
        assert actual[1].value.data == [2]

    @pytest.mark.parametrize(
        'scheduler_conditions, multi_run', [
            (False, False),
            (True, False),
            (True, True)
        ]
    )
    def test_log_multi_calls_single_timestep(self, scheduler_conditions, multi_run):
        lca = pnl.LCAMechanism(
            input_shapes=2,
            leak=0.5,
            threshold=0.515,
            reset_stateful_function_when=pnl.AtTrialStart()
        )
        lca.set_delivery_conditions(pnl.VALUE)
        m0 = pnl.ProcessingMechanism(
            input_shapes=2
        )
        comp = pnl.Composition()
        comp.add_linear_processing_pathway([m0, lca])
        if scheduler_conditions:
            comp.scheduler.add_condition(lca, pnl.AfterNCalls(m0, 2))
        con_with_rpc_pipeline = pnl.Context(rpc_pipeline=Queue(), execution_id=comp)
        pipeline = con_with_rpc_pipeline.rpc_pipeline
        comp.run(inputs={m0: [[1, 0], [1, 0], [1, 0]]}, context=con_with_rpc_pipeline)

        actual = []
        while not pipeline.empty():
            actual.append(pipeline.get())
        integration_end_dict = {i.time: i for i in actual}
        if scheduler_conditions:
            expected_times = ['0:0:1:1', '0:1:1:1', '0:2:1:1']
        else:
            expected_times = ['0:0:0:1', '0:1:0:1', '0:2:0:1']
        assert list(integration_end_dict.keys()) == expected_times
        vals = [i.value.data for i in integration_end_dict.values()]
        # floats in value, so use np.allclose
        np.testing.assert_allclose(vals, [[0.52466739, 0.47533261]] * 3)
        if multi_run:
            comp.run(inputs={m0: [[1, 0], [1, 0], [1, 0]]}, context=con_with_rpc_pipeline)
            actual = []
            while not pipeline.empty():
                actual.append(pipeline.get())
            integration_end_dict.update({i.time: i for i in actual})
            if scheduler_conditions:
                expected_times = ['0:0:1:1', '0:1:1:1', '0:2:1:1', '1:0:1:1', '1:1:1:1', '1:2:1:1']
            else:
                expected_times = ['0:0:0:1', '0:1:0:1', '0:2:0:1', '1:0:0:1', '1:1:0:1', '1:2:0:1']
            assert list(integration_end_dict.keys()) == expected_times
            vals = [i.value.data for i in integration_end_dict.values()]
            # floats in value, so use np.allclose
            np.testing.assert_allclose(vals, [[0.52466739, 0.47533261]] * 6)

class TestFullModels:
    def test_multilayer(self):
        input_layer = pnl.TransferMechanism(name='input_layer',
                                            function=pnl.Logistic,
                                            input_shapes=2)

        hidden_layer_1 = pnl.TransferMechanism(name='hidden_layer_1',
                                               function=pnl.Logistic,
                                               input_shapes=5)

        hidden_layer_2 = pnl.TransferMechanism(name='hidden_layer_2',
                                               function=pnl.Logistic,
                                               input_shapes=4)

        output_layer = pnl.TransferMechanism(name='output_layer',
                                             function=pnl.Logistic,
                                             input_shapes=3)

        input_weights_matrix = (np.arange(2 * 5).reshape((2, 5)) + 1) / (2 * 5)
        middle_weights_matrix = (np.arange(5 * 4).reshape((5, 4)) + 1) / (5 * 4)
        output_weights_matrix = (np.arange(4 * 3).reshape((4, 3)) + 1) / (4 * 3)

        # This projection will be used by the process below by referencing it in the process' pathway;
        #    note: sender and receiver args don't need to be specified
        input_weights = pnl.MappingProjection(
            name='Input Weights',
            matrix=input_weights_matrix,
        )

        # This projection will be used by the process below by assigning its sender and receiver args
        #    to mechanismss in the pathway
        middle_weights = pnl.MappingProjection(
            name='Middle Weights',
            sender=hidden_layer_1,
            receiver=hidden_layer_2,
            matrix=middle_weights_matrix,
        )

        # Commented lines in this projection illustrate variety of ways in which matrix and learning signals can be specified
        output_weights = pnl.MappingProjection(
            name='Output Weights',
            sender=hidden_layer_2,
            receiver=output_layer,
            matrix=output_weights_matrix,
        )

        comp = pnl.Composition(name='multilayer')

        p = [input_layer, input_weights, hidden_layer_1, middle_weights, hidden_layer_2, output_weights, output_layer]
        backprop_pathway = comp.add_backpropagation_learning_pathway(
            pathway=p,
            # loss_spec=pnl.Loss.SSE,
            loss_spec=pnl.Loss.L0,
            learning_rate=1.
        )

        input_dictionary = {backprop_pathway.target: [[0., 0., 1.]],
                            input_layer: [[-1., 30.]]}

        middle_weights.set_delivery_conditions(('mod_matrix', pnl.PROCESSING))
        con_with_rpc_pipeline = pnl.Context(rpc_pipeline=Queue(), execution_id=comp)
        pipeline = con_with_rpc_pipeline.rpc_pipeline

        comp.learn(inputs=input_dictionary,
                   num_trials=10,
                   context=con_with_rpc_pipeline)

        expected_log_val = np.array(
            [
                ['multilayer'],
                [[
                    [[0], [0], [0], [0], [0], [0], [0], [0], [0], [0]],
                    [[0], [1], [2], [3], [4], [5], [6], [7], [8], [9]],
                    [[0], [0], [0], [0], [0], [0], [0], [0], [0], [0]],
                    [[2], [2], [2], [2], [2], [2], [2], [2], [2], [2]],
                    [[[0.05, 0.1, 0.15, 0.2],
                      [0.25, 0.3, 0.35, 0.4],
                      [0.45, 0.5, 0.55, 0.6],
                      [0.65, 0.7, 0.75, 0.8],
                      [0.85, 0.9, 0.95, 1.]],
                     [[0.04789907, 0.09413833, 0.14134241, 0.18938924],
                      [0.24780811, 0.29388455, 0.34096758, 0.38892985],
                      [0.44772121, 0.49364209, 0.54060947, 0.58849095],
                      [0.64763875, 0.69341202, 0.74026967, 0.78807449],
                      [0.84756101, 0.89319513, 0.93994932, 0.98768187]],
                     [[0.04738148, 0.08891106, 0.13248753, 0.177898],
                      [0.24726841, 0.28843403, 0.33173452, 0.37694783],
                      [0.44716034, 0.48797777, 0.53101423, 0.57603893],
                      [0.64705774, 0.6875443, 0.73032986, 0.77517531],
                      [0.84696096, 0.88713512, 0.92968378, 0.97435998]],
                     [[0.04937771, 0.08530344, 0.12439361, 0.16640433],
                      [0.24934878, 0.28467436, 0.32329947, 0.36496974],
                      [0.44932147, 0.48407216, 0.52225175, 0.56359587],
                      [0.64929589, 0.68349948, 0.72125508, 0.76228876],
                      [0.84927212, 0.88295836, 0.92031297, 0.96105307]],
                     [[0.05440291, 0.08430585, 0.1183739, 0.15641064],
                      [0.25458348, 0.28363519, 0.3170288, 0.35455942],
                      [0.45475764, 0.48299299, 0.51573974, 0.55278488],
                      [0.65492462, 0.68238209, 0.7145124, 0.75109483],
                      [0.85508376, 0.88180465, 0.91335119, 0.94949538]],
                     [[0.06177218, 0.0860581, 0.11525064, 0.14926369],
                      [0.26225812, 0.28546004, 0.31377611, 0.34711631],
                      [0.46272625, 0.48488774, 0.51236246, 0.54505667],
                      [0.66317453, 0.68434373, 0.7110159, 0.74309381],
                      [0.86360121, 0.88382991, 0.9097413, 0.94123489]],
                     [[0.06989398, 0.08959148, 0.11465594, 0.14513241],
                      [0.27071639, 0.2891398, 0.31315677, 0.34281389],
                      [0.47150846, 0.48870843, 0.5117194, 0.54058946],
                      [0.67226675, 0.68829929, 0.71035014, 0.73846891],
                      [0.87298831, 0.88791376, 0.90905395, 0.93646]],
                     [[0.07750784, 0.09371987, 0.11555569, 0.143181],
                      [0.27864693, 0.29343991, 0.31409396, 0.3407813],
                      [0.47974374, 0.49317377, 0.5126926, 0.53847878],
                      [0.68079346, 0.69292265, 0.71135777, 0.73628353],
                      [0.88179203, 0.89268732, 0.91009431, 0.93420362]],
                     [[0.0841765, 0.09776672, 0.11711835, 0.14249779],
                      [0.28559463, 0.29765609, 0.31572199, 0.34006951],
                      [0.48695967, 0.49755273, 0.51438349, 0.5377395],
                      [0.68826567, 0.69745713, 0.71310872, 0.735518],
                      [0.88950757, 0.89736946, 0.91190228, 0.93341316]],
                     [[0.08992499, 0.10150104, 0.11891032, 0.14250149],
                      [0.29158517, 0.30154765, 0.31758943, 0.34007336],
                      [0.49318268, 0.50159531, 0.51632339, 0.5377435],
                      [0.69471052, 0.70164382, 0.71511777, 0.73552215],
                      [0.8961628, 0.90169281, 0.91397691, 0.93341744]]]
                ]]
            ],
            dtype=object
        )
        actual = []
        while not pipeline.empty():
            actual.append(pipeline.get())
        log_val = [np.ndarray(shape=np.array(i.value.shape), buffer=np.array(i.value.data)) for i in actual]
        assert all([True if i.context == 'multilayer' else False for i in actual])
        np.testing.assert_allclose(log_val, expected_log_val[1][0][4])

        # Test Programatic logging
        hidden_layer_2.log._deliver_values(pnl.VALUE, con_with_rpc_pipeline)
        actual = []
        while not pipeline.empty():
            actual.append(pipeline.get())
        log_val = [np.ndarray(shape=np.array(actual[0].value.shape), buffer=np.array(actual[0].value.data))]
        expected_log_val = np.array(
            [
                ['multilayer'],
                [[
                    [[1]],
                    [[0]],
                    [[0]],
                    [[0]],
                    [[[0.8565238418942037, 0.8601053239957609, 0.8662098921116546, 0.8746933736954071]]]
                ]]
            ],
            dtype=object
        )
        assert actual[0].context == 'multilayer'
        assert actual[0].time == '1:0:0:0'
        np.testing.assert_allclose(
            expected_log_val[1][0][4],
            log_val
        )

        # Clear log and test with logging of weights set to LEARNING for another 5 trials of learning
        middle_weights.set_delivery_conditions(('mod_matrix', pnl.LEARNING))
        comp.learn(
            num_trials=5,
            inputs=input_dictionary,
            context=con_with_rpc_pipeline
        )
        actual = []
        while not pipeline.empty():
            actual.append(pipeline.get())
        assert all([True if i.context == 'multilayer' else False for i in actual])
        matrices = [i for i in actual if i.parameterName == 'matrix']
        log_val = [np.ndarray(shape=np.array(i.value.shape), buffer=np.array(i.value.data)) for i in matrices]
        expected_log_val = np.array(
            [
                ['multilayer'],
                [[
                    [[1], [1], [1], [1], [1]],  # RUN
                    # [[0], [1], [2], [3], [4]],  # TRIAL
                    [[0], [1], [2], [3], [4]],  # TRIAL
                    [[1], [1], [1], [1], [1]],  # PASS
                    # [[0], [0], [0], [0], [0]],  # PASS
                    # [[1], [1], [1], [1], [1]],  # TIME_STEP
                    [[0], [0], [0], [0], [0]],  # TIME_STEP
                    [[[0.09925812411381937, 0.1079522130303428, 0.12252820028789306, 0.14345816973727732],
                      [0.30131473371328343, 0.30827285172236585, 0.3213609999139731, 0.3410707131678078],
                      [0.5032924245149345, 0.5085833053183328, 0.5202423523987703, 0.5387798509126243],
                      [0.70518251216691, 0.7088822116145151, 0.7191771716324874, 0.7365956448426355],
                      [0.9069777724600303, 0.9091682860319945, 0.9181692763668221, 0.93452610920817]],
                     [[0.103113468050986, 0.11073719161508278, 0.12424368674464399, 0.14415219181047598],
                      [0.3053351724284921, 0.3111770895557729, 0.3231499474835138, 0.341794454877438],
                      [0.5074709829757806, 0.5116017638574931, 0.5221016574478528, 0.5395320566440044],
                      [0.7095115080472698, 0.7120093413898914, 0.7211034158081356, 0.7373749316571768],
                      [0.9114489813353512, 0.9123981459792809, 0.9201588001021687, 0.935330996581107]],
                     [[0.10656261740658036, 0.11328192907953168, 0.12587702586370172, 0.14490737831188183],
                      [0.30893272045369513, 0.31383131362555394, 0.32485356055342113, 0.3425821330631872],
                      [0.5112105492674988, 0.5143607671543178, 0.5238725230390068, 0.5403508295336265],
                      [0.7133860755337162, 0.7148679468096026, 0.7229382109974996, 0.7382232628724675],
                      [0.9154510531345043, 0.9153508224199809, 0.9220539747533424, 0.936207244690072]],
                     [[0.10967776822419642, 0.11562091141141007, 0.12742795007904037, 0.14569308665620523],
                      [0.3121824816018084, 0.316271366885665, 0.3264715025259811, 0.34340179304134666],
                      [0.5145890402653069, 0.5168974760377518, 0.5255545550838675, 0.5412029579613059],
                      [0.7168868378231593, 0.7174964619674593, 0.7246811176253708, 0.7391062307617761],
                      [0.9190671994078436, 0.9180659725806082, 0.923854327015523, 0.9371193149131859]],
                     [[0.11251466428344682, 0.11778293740676549, 0.12890014813698167, 0.14649079441816393],
                      [0.31514245505635713, 0.3185271913574249, 0.328007571201157, 0.3442341089776976],
                      [0.5176666356203712, 0.5192429413004418, 0.5271516632648602, 0.5420683480396268],
                      [0.7200760707077265, 0.7199270072739019, 0.7263361597421493, 0.7400030122347587],
                      [0.922361699102421, 0.9205767427437028, 0.9255639970037588, 0.9380456963960624]]]
                ]]
            ],
            dtype=object
        )

        assert [i.time for i in matrices] == ['1:0:1:0', '1:1:1:0', '1:2:1:0', '1:3:1:0', '1:4:1:0']
        np.testing.assert_allclose(
            expected_log_val[1][0][4],
            log_val
        )



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ********************************************* PNL LLVM helpers **************************************************************

import ast
from contextlib import contextmanager
import warnings

from llvmlite import ir

from .debug import debug_env


@contextmanager
def for_loop(builder, start, stop, inc, id):
    # Initialize index variable
    assert start.type is stop.type
    index_var = builder.alloca(stop.type, name=id + "_index_var_loc")
    builder.store(start, index_var)

    # basic blocks
    cond_block = builder.append_basic_block(id + "-cond-bb")
    out_block = None

    # Loop condition
    builder.branch(cond_block)
    with builder.goto_block(cond_block):
        tmp = builder.load(index_var, name=id + "_cond_index_var")
        cond = builder.icmp_signed("<", tmp, stop, name=id + "_loop_cond")

        # Loop body
        with builder.if_then(cond, likely=True):
            index = builder.load(index_var, name=id + "_loop_index_var")

            yield (builder, index)

            index = builder.add(index, inc, name=id + "_index_var_inc")
            builder.store(index, index_var)
            builder.branch(cond_block)

        out_block = builder.block

    builder.position_at_end(out_block)


def for_loop_zero_inc(builder, stop, id):
    start = stop.type(0)
    inc = stop.type(1)
    return for_loop(builder, start, stop, inc, id)


def array_ptr_loop(builder, array, id):
    # Assume we'll never have more than 4GB arrays
    stop = ir.IntType(32)(array.type.pointee.count)
    return for_loop_zero_inc(builder, stop, id)

def memcpy(builder, dst, src):

    bool_ty = ir.IntType(1)
    char_ptr_ty = ir.IntType(8).as_pointer()
    ptr_src = builder.bitcast(src, char_ptr_ty)
    ptr_dst = builder.bitcast(dst, char_ptr_ty)

    obj_size_ty = ir.FunctionType(ir.IntType(64), [char_ptr_ty, bool_ty, bool_ty, bool_ty])
    obj_size_f = builder.function.module.declare_intrinsic("llvm.objectsize.i64", [], obj_size_ty)
    # the params are: obj pointer, 0 on unknown size, NULL is unknown, size at runtime
    obj_size = builder.call(obj_size_f, [ptr_dst, bool_ty(1), bool_ty(0), bool_ty(0)])

    if "unaligned_copy" in debug_env:
        memcpy_ty = ir.FunctionType(ir.VoidType(), [char_ptr_ty, char_ptr_ty, obj_size.type, bool_ty])
        memcpy_f = builder.function.module.declare_intrinsic("llvm.memcpy", [], memcpy_ty)
        builder.call(memcpy_f, [ptr_dst, ptr_src, obj_size, bool_ty(0)])
    else:
        int_ty = ir.IntType(32)
        int_ptr_ty = int_ty.as_pointer()
        obj_size = builder.add(obj_size, obj_size.type((int_ty.width // 8) - 1))
        obj_size = builder.udiv(obj_size, obj_size.type(int_ty.width // 8))
        ptr_src = builder.bitcast(src, int_ptr_ty)
        ptr_dst = builder.bitcast(dst, int_ptr_ty)

        with for_loop_zero_inc(builder, obj_size, id="memcopy_loop") as (b, idx):
            src = b.gep(ptr_src, [idx])
            dst = b.gep(ptr_dst, [idx])
            b.store(b.load(src), dst)

    return builder


def fclamp(builder, val, min_val, max_val):
    min_val = min_val if isinstance(min_val, ir.Value) else val.type(min_val)
    max_val = max_val if isinstance(max_val, ir.Value) else val.type(max_val)

    cond = builder.fcmp_unordered("<", val, min_val)
    tmp = builder.select(cond, min_val, val)
    cond = builder.fcmp_unordered(">", tmp, max_val)
    return builder.select(cond, max_val, tmp)


def uint_min(builder, val, other):
    other = other if isinstance(other, ir.Value) else val.type(other)

    cond = builder.icmp_unsigned("<=", val, other)
    return builder.select(cond, val, other)


def get_param_ptr(builder, component, params_ptr, param_name):
    # check if the passed location matches expected size
    assert len(params_ptr.type.pointee) == len(component.llvm_param_ids)

    idx = ir.IntType(32)(component.llvm_param_ids.index(param_name))
    return builder.gep(params_ptr, [ir.IntType(32)(0), idx],
                       name="ptr_param_{}_{}".format(param_name, component.name))


def get_state_ptr(builder, component, state_ptr, stateful_name, hist_idx=0):
    # check if the passed location matches expected size
    assert len(state_ptr.type.pointee) == len(component.llvm_state_ids)

    idx = ir.IntType(32)(component.llvm_state_ids.index(stateful_name))
    ptr = builder.gep(state_ptr, [ir.IntType(32)(0), idx],
                      name="ptr_state_{}_{}".format(stateful_name,
                                                    component.name))
    # The first dimension of arrays is history
    if hist_idx is not None and isinstance(ptr.type.pointee, ir.ArrayType):
        assert len(ptr.type.pointee) > hist_idx, \
            "History not available: {} ({})".format(ptr.type.pointee, hist_idx)
        ptr = builder.gep(state_ptr, [ir.IntType(32)(0), idx,
                                      ir.IntType(32)(hist_idx)],
                          name="ptr_state_{}_{}_hist{}".format(stateful_name,
                                                               component.name,
                                                               hist_idx))
    return ptr


def get_state_space(builder, component, state_ptr, name):
    val_ptr = get_state_ptr(builder, component, state_ptr, name, None)
    for i in range(len(val_ptr.type.pointee) - 1, 0, -1):
        dest_ptr = get_state_ptr(builder, component, state_ptr, name, i)
        src_ptr = get_state_ptr(builder, component, state_ptr, name, i - 1)
        builder.store(builder.load(src_ptr), dest_ptr)

    return get_state_ptr(builder, component, state_ptr, name)


def unwrap_2d_array(builder, element):
    if isinstance(element.type.pointee, ir.ArrayType) and isinstance(element.type.pointee.element, ir.ArrayType):
        assert element.type.pointee.count == 1
        return builder.gep(element, [ir.IntType(32)(0), ir.IntType(32)(0)])
    return element


def load_extract_scalar_array_one(builder, ptr):
    val = builder.load(ptr)
    if isinstance(val.type, ir.ArrayType) and val.type.count == 1:
        val = builder.extract_value(val, [0])
    return val


def umul_lo_hi(builder, a, b):
    assert a.type.width == b.type.width

    a_val = builder.zext(a, ir.IntType(a.type.width * 2))
    b_val = builder.zext(b, ir.IntType(b.type.width * 2))
    res = builder.mul(a_val, b_val)

    lo = builder.trunc(res, a.type)
    hi = builder.lshr(res, res.type(a.type.width))
    hi = builder.trunc(hi, a.type)
    return lo, hi


def fneg(builder, val, name=""):
    return builder.fsub(val.type(-0.0), val, name)


def exp(ctx, builder, x):
    exp_f = ctx.get_builtin("exp", [x.type])
    return builder.call(exp_f, [x])

def log(ctx, builder, x):
    log_f = ctx.get_builtin("log", [x.type])
    return builder.call(log_f, [x])

def log1p(ctx, builder, x):
    log_f = ctx.get_builtin("log", [x.type])
    x1p = builder.fadd(x, x.type(1))
    return builder.call(log_f, [x1p])

def sqrt(ctx, builder, x):
    sqrt_f = ctx.get_builtin("sqrt", [x.type])
    return builder.call(sqrt_f, [x])

def tanh(ctx, builder, x):
    tanh_f = ctx.get_builtin("tanh", [x.type])
    return builder.call(tanh_f, [x])


def coth(ctx, builder, x):
    coth_f = ctx.get_builtin("coth", [x.type])
    return builder.call(coth_f, [x])


def csch(ctx, builder, x):
    csch_f = ctx.get_builtin("csch", [x.type])
    return builder.call(csch_f, [x])


def is_close(ctx, builder, val1, val2, rtol=1e-05, atol=1e-08):
    assert val1.type == val2.type
    is_close_f = ctx.get_builtin("is_close_{}".format(val1.type))
    rtol_val = val1.type(rtol)
    atol_val = val1.type(atol)
    return builder.call(is_close_f, [val1, val2, rtol_val, atol_val])


def all_close(ctx, builder, arr1, arr2, rtol=1e-05, atol=1e-08):
    assert arr1.type == arr2.type
    all_ptr = builder.alloca(ir.IntType(1), name="all_close_slot")
    builder.store(all_ptr.type.pointee(1), all_ptr)
    with array_ptr_loop(builder, arr1, "all_close") as (b1, idx):
        val1_ptr = b1.gep(arr1, [idx.type(0), idx])
        val2_ptr = b1.gep(arr2, [idx.type(0), idx])
        val1 = b1.load(val1_ptr)
        val2 = b1.load(val2_ptr)
        res_close = is_close(ctx, b1, val1, val2, rtol, atol)

        all_val = b1.load(all_ptr)
        all_val = b1.and_(all_val, res_close)
        b1.store(all_val, all_ptr)

    return builder.load(all_ptr)


def create_sample(builder, allocation, search_space, idx):
    # Construct allocation corresponding to this index
    for i in reversed(range(len(search_space.type.pointee))):
        slot_ptr = builder.gep(allocation, [idx.type(0), idx.type(i)])

        dim_ptr = builder.gep(search_space, [idx.type(0), idx.type(i)])
        # Iterators store {start, step, num}
        if isinstance(dim_ptr.type.pointee,  ir.LiteralStructType):
            iter_val = builder.load(dim_ptr)
            dim_start = builder.extract_value(iter_val, 0)
            dim_step = builder.extract_value(iter_val, 1)
            dim_size = builder.extract_value(iter_val, 2)
            dim_idx = builder.urem(idx, dim_size)
            val = builder.uitofp(dim_idx, dim_step.type)
            val = builder.fmul(val, dim_step)
            val = builder.fadd(val, dim_start)
        elif isinstance(dim_ptr.type.pointee,  ir.ArrayType):
            # Otherwise it's just an array
            dim_size = idx.type(len(dim_ptr.type.pointee))
            dim_idx = builder.urem(idx, dim_size)
            val_ptr = builder.gep(dim_ptr, [idx.type(0), dim_idx])
            val = builder.load(val_ptr)
        else:
            assert False, "Unknown dimension type: {}".format(dim_ptr.type)

        idx = builder.udiv(idx, dim_size)

        builder.store(val, slot_ptr)


def convert_type(builder, val, t):
    assert isinstance(t, ir.Type)
    if val.type == t:
        return val

    if is_floating_point(val) and is_boolean(t):
        # convert any scalar to bool by comparing to 0
        # Python converts both 0.0 and -0.0 to False
        return builder.fcmp_unordered("!=", val, val.type(0.0))

    if is_boolean(val) and is_floating_point(t):
        # float(True) == 1.0, float(False) == 0.0
        return builder.select(val, t(1.0), t(0.0))

    if is_integer(val) and is_integer(t):
        if val.type.width > t.width:
            return builder.trunc(val, t)
        elif val.type.width < t.width:
            # Python integers are signed
            return builder.zext(val, t)
        else:
            assert False, "Unknown integer conversion: {} -> {}".format(val.type, t)

    if is_integer(val) and is_floating_point(t):
        # Python integers are signed
        return builder.sitofp(val, t)

    if is_floating_point(val) and is_integer(t):
        # Python integers are signed
        return builder.fptosi(val, t)

    if is_floating_point(val) and is_floating_point(t):
        if isinstance(val.type, ir.HalfType) or isinstance(t, ir.DoubleType):
            return builder.fpext(val, t)
        elif isinstance(val.type, ir.DoubleType) or isinstance(t, ir.HalfType):
            # FIXME: Direct conversion from double to half needs a runtime
            #        function (__truncdfhf2). llvmlite MCJIT fails to provide
            #        it and instead generates invocation of a NULL pointer.
            #        Use double conversion (double->float->half) instead.
            #        Both steps can be done in one CPU instruction,
            #        but the result can be slightly different
            #        see: https://github.com/numba/llvmlite/issues/834
            if isinstance(val.type, ir.DoubleType) and isinstance(t, ir.HalfType):
                val = builder.fptrunc(val, ir.FloatType())
            return builder.fptrunc(val, t)
        else:
            assert False, "Unknown float conversion: {} -> {}".format(val.type, t)

    assert False, "Unknown type conversion: {} -> {}".format(val.type, t)


def is_pointer(x):
    type_t = getattr(x, "type", x)
    assert isinstance(type_t, ir.Type)
    return isinstance(type_t, ir.PointerType)

def is_floating_point(x):
    type_t = getattr(x, "type", x)
    # dereference pointer
    if is_pointer(type_t):
        type_t = type_t.pointee
    return isinstance(type_t, (ir.DoubleType, ir.FloatType, ir.HalfType))

def is_integer(x):
    type_t = getattr(x, "type", x)
    # dereference pointer
    if is_pointer(type_t):
        type_t = type_t.pointee
    return isinstance(type_t, ir.IntType)

def is_scalar(x):
    return is_integer(x) or is_floating_point(x)

def is_vector(x):
    type_t = getattr(x, "type", x)
    if is_pointer(type_t):
        type_t = type_t.pointee
    return isinstance(type_t, ir.ArrayType) and is_scalar(type_t.element)

def is_2d_matrix(x):
    type_t = getattr(x, "type", x)
    if is_pointer(type_t):
        type_t = type_t.pointee
    return isinstance(type_t, ir.ArrayType) and is_vector(type_t.element)

def is_boolean(x):
    type_t = getattr(x, "type", x)
    if is_pointer(type_t):
        type_t = type_t.pointee
    return isinstance(type_t, ir.IntType) and type_t.width == 1

def get_array_shape(x):
    x_ty = x.type
    if is_pointer(x):
        x_ty = x_ty.pointee

    assert isinstance(x_ty, ir.ArrayType), f"Tried to get shape of non-array type: {x_ty}"
    dimensions = []
    while hasattr(x_ty, "count"):
        dimensions.append(x_ty.count)
        x_ty = x_ty.element

    return dimensions

def array_from_shape(shape, element_ty):
    array_ty = element_ty
    for dim in reversed(shape):
        array_ty = ir.ArrayType(array_ty, dim)
    return array_ty

@contextmanager
def recursive_iterate_arrays(ctx, builder, *args, loop_id="recursive_iteration"):
    """Recursively iterates over all elements in scalar arrays of the same shape"""

    assert len(args) > 0, "Need at least one array to iterate over!"
    assert all(isinstance(arr.type.pointee, ir.ArrayType) for arr in args), "Can only iterate over arrays!"

    u = args[0]
    assert all(len(u.type.pointee) == len(v.type.pointee) for v in args), "Tried to iterate over differing lengths!"

    with array_ptr_loop(builder, u, loop_id) as (b, idx):
        arg_ptrs = tuple(b.gep(arr, [ctx.int32_ty(0), idx]) for arr in args)
        if is_scalar(arg_ptrs[0]):
            yield (b, *arg_ptrs)
        else:
            with recursive_iterate_arrays(ctx, b, *arg_ptrs) as (b, *nested_args):
                yield (b, *nested_args)

def printf(ctx, builder, fmt, *args, tags:set):

    tags = frozenset(tags)
    user_tags = frozenset(ast.literal_eval(debug_env.get("printf_tags", "[]")))
    if "all" not in user_tags and "always" not in tags and not tags.intersection(user_tags):
        return

    # Set up the formatting string as global symbol
    int8 = ir.IntType(8)
    fmt_data = bytearray((fmt + "\0").encode("utf8"))
    fmt_ty = ir.ArrayType(int8, len(fmt_data))

    ir_module = builder.function.module
    global_fmt = ir.GlobalVariable(ir_module,
                                   fmt_ty,
                                   name="printf_fmt_" + str(len(ir_module.globals)))
    global_fmt.linkage = "internal"
    global_fmt.global_constant = True
    global_fmt.initializer = fmt_ty(fmt_data)

    printf_ty = ir.FunctionType(ir.IntType(32), [ir.IntType(8).as_pointer()], var_arg=True)
    get_printf_addr_f = ctx.get_builtin("get_printf_address", [])
    printf_address = builder.call(get_printf_addr_f, [])

    printf_is_not_null = builder.icmp_unsigned("!=", printf_address, printf_address.type(0))
    with builder.if_then(printf_is_not_null, likely=True):
        printf_f = builder.inttoptr(printf_address, printf_ty.as_pointer())

        fmt_ptr = builder.gep(global_fmt, [ir.IntType(32)(0), ir.IntType(32)(0)])
        conv_args = [builder.fpext(a, ir.DoubleType()) if is_floating_point(a) else a for a in args]
        builder.call(printf_f, [fmt_ptr] + conv_args)


def printf_float_array(ctx, builder, array, prefix="", suffix="\n", *, tags:set):
    printf(ctx, builder, prefix, tags=tags)

    with array_ptr_loop(builder, array, "print_array_loop") as (b1, i):
        printf(ctx, b1, "%lf ", b1.load(b1.gep(array, [i.type(0), i])), tags=tags)

    printf(ctx, builder, suffix, tags=tags)


def printf_float_matrix(ctx, builder, matrix, prefix="", suffix="\n", *, tags:set):
    printf(ctx, builder, prefix, tags=tags)
    with array_ptr_loop(builder, matrix, "print_row_loop") as (b1, i):
        row = b1.gep(matrix, [i.type(0), i])
        printf_float_array(ctx, b1, row, suffix="\n", tags=tags)

    printf(ctx, builder, suffix, tags=tags)



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# *********************************************** Pathway **************************************************************

"""

Related
-------

* `NodeRoles <NodeRole>`
* `PathwayRoles <PathwayRole>`
* `Composition`

Contents
--------

  * `Pathway_Overview`
  * `Pathway_Creation`
      - `Pathway_Template`
      - `Pathway_Assignment_to_Composition`
      - `Pathway_Name`
      - `Pathway_Specification`
          - `Pathway_Specification_Formats`
          - `Pathway_Specification_Projections`
          - `Pathway_Specification_Multiple`
      - `Composition_Add_Nested`
  * `Pathway_Structure`
  * `Pathway_Execution`
  * `Pathway_Class_Reference`

.. _Pathway_Overview:

Overview
--------

A Pathway is a sequence of `Nodes <Composition_Nodes>` and `Projections <Projection>`. Generally, Pathways are assigned
to a `Compositions`, but a Pathway object can also be created on its and used as a template for specifying a Pathway for
a Composition, as described below (see `Pathways  <Composition_Pathways>` for additional information about Pathways in
Compositions).

.. _Pathway_Creation:

Creating a Pathway
------------------

Pathway objects are created in one of two ways, either using the constructor to create a `template <Pathway_Template>`,
or automatically when a Pathway is `assigned to a Composition <Pathway_Assignment_to_Composition>`.

.. _Pathway_Template:

*Pathway as a Template*
~~~~~~~~~~~~~~~~~~~~~~~

A Pathway created on its own, using its constructor, is a **template**, that can be used to `specify a Pathway
<Pathway_Specification>` for one or more Compositions, as described `below <Pathway_Assignment_to_Composition>`;
however, it cannot be executed on its own.  When a Pathway object is used to assign a Pathway to a Composition,
its `pathway <Pathway.pathway>` attribute, and its `name <Pathway.name>` if that is not otherwise specified (see
`below <Pathway_Name>`), are used as the specification to create a new Pathway that is assigned to the Composition,
and the template remains unassigned.

.. _Pathway_Assignment_to_Composition:

*Assigning Pathways to a Composition*
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Pathways can be assigned to a `Composition` using the **pathways** argument of the Composition's constructor, or one of
its `Pathway addition methods <Composition_Pathway_Addition_Methods>`.  A Pathway object can be used, along with any
other form of `Pathway specification <Pathway_Specification>` for those arguments.  However, the Pathway object itself
is not assigned to the Composition; rather, it serves as a `template  <Pathway_Template>` for a new Pathway object
that is automatically created and assigned to the Composition.  The Composition's `Pathway addition methods
<Composition_Pathway_Addition_Methods>` return a Pathway object for each Pathway created and assigned to theComposition.
All of the Pathways assigned to a Composition are listed in its `pathways <Composition.pathways>` attribute.

.. _Pathway_Name:

*Pathway Name*
~~~~~~~~~~~~~~

If the **name** argument of the Pathway's constructor is used to assign it a name, this is used as the name of the
Pathway created when it is assigned to a Composition in its constructor, or using its `add_pathways
<Composition.add_pathways>` method.  This is also the case if one of the Composition's other `Pathway addition methods
<Composition_Pathway_Addition_Methods>` is used, as long as the **name** argument of those methods is not specified.
However, if the **name** argument is specified in those methods, or `Pathway specification dictionary
<Pathway_Specification_Dictionary>` is used to specify the Pathway's name, that takes precedence over, and replaces
one specified in the Pathway `template's <Pathway_Template>` `name <Pathway.name>` attribute.


.. _Pathway_Specification:

*Pathway Specification*
~~~~~~~~~~~~~~~~~~~~~~~

A Pathway is specified as a list, each element of which is either a `Node <Composition_Nodes>` or set of Nodes,
possibly intercolated with specifications of `Projections <Projection>` between them.  The specification of each
is described in the two sections below.

.. _Pathway_Specification_Nodes:

*Node Specifications*
^^^^^^^^^^^^^^^^^^^^^

`Nodes <Composition_Nodes>` can be either a `Mechanism`, a `Composition`, or a tuple (Mechanism or Composition,
`NodeRoles <NodeRole>`) that can be used to assign `required_roles` to the Nodes in the Composition (see
`Composition_Nodes` for additional details). The Node(s) specified in each entry of the list project to the
Node(s) specified in the next entry, with the exception of a `ControlMechanism`.

**ControlMechanisms**. Any ControlMechanisms specified in a Pathway are automatically added as `Nodes
<Composition_Nodes>` of the Composition. If a ControlMechanism follows another Node that is not a ControlMechanism,
then a `MappingProjection` is created to it from that Node (as described `below <Pathway_Projections>`). However, *no*
MappingProjections are created from the ControlMechanism to any that *follow* it in the Pathway specification.  This is
because the efferent Projections of a ControlMechanism can only be `ModulatoryProjection`\\s (see `
<ControlMechanism_ControlSignals>` for specification), and not MappingProjections.  Instead, if the ControlMechanism is
the first Node in the Pathway specification, the Node that follows it is treated as an `INPUT` Node of the Composition
(unless it receives a Projection from another Pathway);  if it is the last Node in the Pathway specification, the Node
that precedes it is treated as an `OUTPUT` Node of the Composition (unless it sends a Projection to another Pathway);
otherwise, a MappingProjection is created from the `primary OutputPort <OutputPort_Primary>` of the Node immediately
preceding the ControlMechanism to the `primary InputPort <InputPort_Primary>` of the Node immediately following it.

**Use of sets**.  A set of Nodes can be used to specify multiple Nodes, all of which should receive Projections
from the same Node(s) in the preceding entry in the Pathway specification and send Projections to the same Node(s) in
the following entry (see `below <Pathway_Projections>` for additional details).

    .. _Pathway_Projection_List_Note:

    .. note::
       Only a *set* can be used to specify multiple Nodes for a given entry in a Pathway; a *list* can *not* be used
       for this purpose, as a list containing Nodes is always interpreted as a Pathway. If a list *is* included in a
       Pathway specification, then it and all other entries are considered as separate, parallel Pathways (see
       example *vii* in the `figure <Pathway_Figure>` below).

.. _Pathway_Specification_Projections:

*Projection Specifications*
^^^^^^^^^^^^^^^^^^^^^^^^^^^

Where no Projections are specified between entries in the list, default Projections are created (using a
`FULL_CONNECTIVITY_MATRIX`, or the Pathway's `default_projection <Pathway.default_projection_matrix>` if specified)
from each Node in the first entry, as the sender(s), to each Node in the second, as receiver(s) (described further
`below <Pathway_Projections>`).  Projections between Nodes in the two entries can also be specified explicitly, by
intercolating a Projection or set of Projections between the two entries in the list. If the sender and receiver are
both a single Mechanism, then a single `MappingProjection` can be `specified<MappingProjection_Creation>` between
them.  The same applies if the sender is a `Composition` with a single `OUTPUT <NodeRole.OUTPUT>` Node and/or the
receiver is a `Composition` with a single `INPUT <NodeRole.INPUT>` Node.  If either is a set of Nodes, or is a
`nested Composition <Composition_Nested>` with more than one `INPUT <NodeRole.INPUT>` or `OUTPUT <NodeRole.OUTPUT>`
Node, respectively, then a collection of Projections can be specified between any or all pairs of the Nodes in the
set(s) and/or nested Composition(s), using either a set or list of Projections (order of specification does not matter
whether a set or a list is used). The collection can contain `MappingProjections <MappingProjection>` between specific
pairs of Nodes and/or a single default specification (either a `matrix <MappingProjection.matrix>` specification or a
MappingProjection without any `sender <MappingProjection.sender>` or `receiver <MappingProjection.receiver>`
specified; see MappingProject  MappingProjection_Matrix_Specification
).

    .. _Pathway_Projection_Matrix_Note:

    .. note::
       If a collection of Projection specifications includes a default matrix specification, then a list must be used
       to specify the collection and *not* a set (since a matrix is unhashable and thus cannot be included in a set).

If a default Projection specification is included in the set, it is used to implement a Projection between any pair
of Nodes for which no MappingProjection is otherwise specified, whether within the collection or on its own; if no
Projections are specified for any individual pairs, a default Projection is created for every pairing of senders and
receivers. If a collection contains Projections for one or more pairs of Nodes, but does not include a default
projection specification, then no Projection is created between any of the other pairings.

If a pair of entries in a pathway has multiple sender and/or receiver Nodes specified (either in a set and/or belonging
to `nested Composition <Composition_Nested>`, and either no Projection(s) or only a default Projection is intercollated
between them, then a default set of Projections is constructed (using the default Projection specification, if provided)
between each pair of sender and receiver Nodes in the set(s) or nested Composition(s), as follows:

.. _Pathway_Projections:

* *One to one* - if both the sender and receiver entries are Mechanisms, or if either is a Composition and the
  sender has a single `OUTPUT <NodeRole.OUTPUT>` Node and the receiver has a single `INPUT <NodeRole.INPUT>`
  Node, then a default `MappingProjection` is created from the `primary OutputPort <OutputPort_Primary>` of the
  sender (or of its sole `OUTPUT <NodeRole.OUTPUT>` Node, if the sender is a Composition) to the `primary InputPort
  <InputPort_Primary>` of the receiver (or of its sole of `INPUT <NodeRole.INPUT>` Node, if the receiver is
  a Composition), and the Projection specification is intercolated between the two entries in the `Pathway`.

* *One to many* - if the sender is either a Mechanism or a Composition with a single `OUTPUT <NodeRole.OUTPUT>` Node,
  but the receiver is either a Composition with more than one `INPUT <NodeRole.INPUT>` Node or a set of Nodes, then
  a `MappingProjection` is created from the `primary OutputPort <OutputPort_Primary>` of the sender Mechanism (or of
  its sole `OUTPUT <NodeRole.OUTPUT>` Node if the sender is a Composition) to the `primary InputPort
  <InputPort_Primary>` of each `INPUT <NodeRole.OUTPUT>` Node of the receiver Composition and/or Mechanism in the
  receiver set, and a set containing the Projections is intercolated between the two entries in the `Pathway`.

* *Many to one* - if the sender is a Composition with more than one `OUTPUT <NodeRole.OUTPUT>` Node or a set of
  Nodes, and the receiver is either a Mechanism or a Composition with a single `INPUT <NodeRole.INPUT>` `OUTPUT
  <NodeRole.OUTPUT>` Node in the Composition or Mechanism in the set of sender(s), to the `primary InputPort
  <InputPort_Primary>` of the receiver Mechanism (or of its sole `INPUT <NodeRole.INPUT>` Node if the receiver is
  a Composition), and a set containing the Projections is intercolated between the two entries in the `Pathway`.

* *Many to many* - if both the sender and receiver entries contain multiple Nodes (i.e., are sets,  and/or the
  the sender is a Composition that has more than one `INPUT <NodeRole.INPUT>` Node and/or the receiver has more
  than one `OUTPUT <NodeRole.OutPUT>` Node), then a Projection is constructed for every pairing of Nodes in the
  sender and receiver entries, using the `primary OutputPort <OutputPort_Primary>` of each sender Node and the
  `primary InputPort <InputPort_Primary>` of each receiver node.

|

  .. _Pathway_Figure:

  .. figure:: _static/Pathways_fig.svg
     :scale: 50%

     **Examples of Pathway specifications** (including in the **pathways** argument of a `Composition`. *i)* Set
     of `Nodes <Composition_Nodes>`: each is treated as a `SINGLETON <NodeRole.SINGLETON>` within a single Pathway.
     *ii)* List of Nodes: forms a sequential Pathway. *iii)* Single Node followed by a set:  one to many mapping.
     *iv)* Set followed by a single Node: many to one mapping. *v)* Set followed by a set: many to many mapping.
     *vi)* Set followed by a list: because there is a list in the specification (``[C,D]``) all other entries are
     also treated as parallel Pathways (see `note <Pathway_Projection_List_Note>` above), so ``A`` and ``B`` in the
     set are `SINGLETON <NodeRole.SINGLETON>`\\s. *vii)* Set of Projections intercolated between two sets of Nodes:
     since the set of Projections does not include any involving ``B`` or ``E`` nor a default Projection specification,
     they are treated as `SINGLETON <NodeRole.SINGLETON>`\\s (compare with *x*). *viii)* Set followed by a Node and
     then a set:  many to one to many mapping. *ix)* Node followed by one that is a `nested Composition
     <Composition_Nested>` then another Node: one to many to one mapping. *x)* Set followed by a list of Projections
     then another set: since the list of Projections contains a default Projection specification (``matrix``)
     Projections are created between all pairings of nodes in the sets that precede and follow the list (compare with
     *vii*); note that the Projections must be specified in a list because the matrix is a list (or array), which
     cannot be included in a set (see `note <Pathway_Projection_Matrix_Note>` above).

     .. technical_note::
        The full code for the examples above can be found in `test_pathways_examples`,
        although some have been graphically rearranged for illustrative purposes.



.. _Pathway_Specification_Formats:

*Pathway Specification Formats*
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The following formats can be used to specify a Pathway in the **pathway** argument of the constructor for
the Pathway, the **pathways** argument of the constructor for a `Composition`, or the corresponding argument
of any of a Composition's `Pathway addition methods <Composition_Pathway_Addition_Methods>`:

    * `Node <Composition_Nodes>`: -- assigns the Node as `SINGLETON <NodeRole.SINGLETON>` in a Pathway.
    ..
    .. _Pathway_Specification_List:

    * **list**: [`Node <Composition_Nodes>`, <`Projection(s) <Projection>`,> `Node <Composition_Nodes>`...] --
      each item of the list must be a `Node <Composition_Nodes>` (i.e., Mechanism or Composition, or a
      (`Mechanism <Mechanism>`, `NodeRoles <NodeRole>`) tuple) or set of Nodes, optionally with a `Projection
      specification <Projection_Specification>`, a (`Projection specification <Projection_Specification>`,
      `feedback specification <Composition_Feedback_Designation>`) tuple, or a set of either interposed between
      a pair of (sets of) Nodes (see `add_linear_processing_pathway <Composition.add_linear_processing_pathway>`
      for additional details).  The list must begin and end with a (set of) Node(s).
    ..
    * **set**: {`Node <Composition_Nodes>`, `Node <Composition_Nodes>`...} --
      each item of the set must be a `Node <Composition_Nodes>` (i.e., Mechanism or Composition, or a
      (`Mechanism <Mechanism>`, `NodeRoles <NodeRole>`) tuple);  each Node is treated as a `SINGLETON
      <NodeRole.SINGLETON>`.  Sets can also be used in a list specification (see above; and see
      `add_linear_processing_pathway <Composition.add_linear_processing_pathway>` for additional details).
    ..
    .. _Pathway_Specification_Tuple:

    * **2 or 3-item tuple**: (Pathway, <learning_function>, <default_matrix_specification>) --
      used to specify a `learning  Pathway <Composition_Learning_Pathway>` and/or a matrix to use for any unspecified
      Projections (overrides default matrix for `MappingProjection`) if a default projection is not otherwise specified
      (see `Pathway_Specification_Projections.  The 1st item of the tuple must be one of the forms of Pathway
      specification described above.  The other items must be a subclass of `LearningFunction` and/or a `matrix
      specification <MappingProjection_Matrix_Specification>`.

.. _Pathway_Specification_Multiple:

*Multiple Pathway Specifications*
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

In addition to the forms of single Pathway specification `above <Pathway_Specification>`, where multiple Pathways
can be specified (e.g., the **pathways** argument of the constructor for a `Composition` or its `add_pathways
<Composition.add_pathways>` method), they can be specified in a list, in which each item of the list can be any of
the forms above, or one of the following:

    * **Pathway** object or constructor: Pathway(pathway=\\ `Pathway specification <Pathway_Specification>`,...).
    ..
    .. _Pathway_Specification_Dictionary:
    * **dict**: {name : Pathway) -- in which **name** is a str and **Pathway** is a Pathway object or constuctor,
      or one of the standard `Pathway specifications <Pathway_Specification>` listed above.

    .. note::
       If any of the following is used to specify the **pathways** argument:
         * a **standalone** `Node <Composition_Nodes>` (i.e., not in a list), \n
         * a **single Node** alone in a list, \n
         * a **set** of Nodes, \n
         * one or more Nodes with any other form of `Pathway specification <Pathway_Specification>` in the list \n
       then each such Node in the list is assigned as a `SINGLETON <NodeRole.SINGLETON>` Node in its own Pathway.
       However, if the list contains only Nodes, then it is treated as a single Pathway (i.e., the list form of
       `Pathway specification <Pathway_Specification>` described above.  Thus:
         **pathway**: NODE -> single pathway \n
         **pathway**: [NODE] -> single pathway \n
         **pathway**: [NODE, NODE...] -> single pathway \n
         **pathway**: [NODE, () or {} or `Pathway`...] -> individual Pathways for each specification.


.. _Pathway_Structure:

Structure
---------

.. _Pathway_Attribute:

A Pathway has the following primary attributes:

* `pathway <Pathway.pathway>` - if the Pathway was created on its own, this contains the specification provided in
  the **pathway** arg of its constructor; that is, depending upon how it was specified, it may or may not contain
  fully constructed `Components <Component>`.  This is passed to the **pathways** argument of a Composition's
  constructor or one of its `pathway addition methods <Composition_Pathway_Addition_Methods>` when the Pathway is used
  in the specification of any of these.  In contrast, when a Pathway is created by a Composition (and assigned to its
  `pathways <Composition.pathways>` attribute), then the actual `Mechanism(s) <Mechanism>` and/or `Composition(s)`
  that comprise `Nodes <Composition_Nodes>`, and the `Projection(s) <Projection>` between them, are listed in the
  Pathway's `pathway <Pathway.pathway>` attribute.

* `composition <Pathway.composition>` - contains the `Composition` that created the Pathway and to which it belongs,
  or None if it is a `template <Pathway_Template>` (i.e., was constructed on its own).

* `roles <Pathway.roles>` and `Node <Composition_Nodes>` attributes - if the Pathway was created by a Composition,
  the `roles <Pathway.roles>` attribute `this lists the `PathwayRoles <PathwayRole>` assigned to it by the Composition
  that correspond to the `NodeRoles <NodeRole`> of its Components, and the `Nodes <Composition_Nodes>` with each of
  those `NodeRoles <NodeRole>` is assigned to a corresponding attribute on the Pathway.  If the Pathway does not belong
  to a Composition (i.e., it is a `template <Pathway_Template>`), then these attributes return None.

* `default_projection_matrix <Pathway.default_projection_matrix>` - matrix used as default for Projections that are
  not explicitly specified and for which no default is otherwise specified (see `Pathway_Specification_Projections`).

* `learning_function <Pathway.learning_function>` - the LearningFunction assigned to the Pathway if it is a
  `learning Pathway <Composition_Learning_Pathway>` that belongs to a Composition; otherwise it is None.

.. _Pathway_Execution:

Execution
---------

A Pathway cannot be executed on its own.  Its Components are executed when the Composition to which it belongs is
executed, by default in the order in which they appear in the `pathway <Pathway.pathway>` attribute;  however, this
can be modified by `Conditions <Condition>` added to the Composition's `scheduler <Composition.scheduler>`.

.. _Pathway_Class_Reference:


Class Reference
---------------

"""
import warnings
from enum import Enum

from psyneulink._typing import Literal

from psyneulink.core.components.shellclasses import Mechanism
from psyneulink.core.compositions.composition import Composition, CompositionError, NodeRole
from psyneulink.core.globals.keywords import \
    ANY, CONTEXT, FEEDBACK, MAYBE, NODE, LEARNING_FUNCTION, OBJECTIVE_MECHANISM, PROJECTION, TARGET_MECHANISM
from psyneulink.core.globals.utilities import is_matrix
from psyneulink.core.globals.registry import register_category

__all__ = [
    'Pathway', 'PathwayRegistry', 'PathwayRole'
]


PathwayRegistry= {}


def _is_pathway_entry_spec(entry, desired_type: Literal['NODE', 'Projection', 'any']):
    """Test whether pathway entry is specified type (NODE or PROJECTION)"""
    from psyneulink.core.components.projections.projection import _is_projection_spec
    node_types = (Mechanism, Composition)
    is_node = is_proj = is_set = False

    if desired_type in {NODE, ANY}:
        is_node = (isinstance(entry, node_types)
                   or (isinstance(entry, tuple)
                       and isinstance(entry[0], node_types)
                       and (isinstance(entry[1], NodeRole) or
                            (isinstance(entry[1], list) and all(isinstance(nr, NodeRole) for nr in entry[1])))))

    if desired_type in {PROJECTION, ANY}:
        is_proj = (_is_projection_spec(entry)
                   or (isinstance(entry, tuple)
                       and _is_projection_spec(entry[0])
                       and entry[1] in {True, FEEDBACK, False, MAYBE})
                   or (isinstance(entry, (set,list))
                   # or (isinstance(entry, set)
                       and all(_is_projection_spec(item) for item in entry)))

    if desired_type in {ANY}:
        is_set = (isinstance(entry, set) and all(_is_node_spec(item) for item in entry))

    if is_node or is_proj or is_set:
        return True
    else:
        return False


def _is_node_spec(value):
    return _is_pathway_entry_spec(value, NODE)


class PathwayRole(Enum):
    """Roles assigned to the `Pathway` of a `Composition`.

    Attributes
    ----------

    ORIGIN
        A `Pathway` that includes an `ORIGIN` `Node <Composition_Nodes>` of the `Composition`.

    INPUT
        A `Pathway` that includes an `INPUT` `Node <Composition_Nodes>` of the `Composition`.

    SINGELTON
        A `Pathway` with a single `Node <Composition_Nodes>` that is a `SINGLETON` of the `Composition`.

    INTERNAL
        A `Pathway` that does not include any `ORIGIN` or `TERMINAL` `Nodes <Composition_Nodes>` of the `Composition`.

    OUTPUT
        A `Pathway` that includes an `OUTPUT` `Node <Composition_Nodes>` of the `Composition`.

    TERMINAL
        A `Pathway` that includes a `TERMINAL` `Node <Composition_Nodes>` of the `Composition`.

    CYCLE
        A `Pathway` that constitutes a `CYCLE` in the `Composition`.

    COMMENT:
    CONTROL
        A `Pathway` that constitutes a `control pathway <Composition_Control_Pathways>` of the `Composition`.
    COMMENT

    LEARNING
        A `Pathway` that constitutes a `learning Pathway <Composition_Learning_Pathway>` of the `Composition`.

    """
    ORIGIN = 0
    INPUT = 1
    SINGLETON = 2
    INTERNAL = 3
    OUTPUT = 4
    TERMINAL = 5
    CYCLE = 6
    CONTROL = 7
    LEARNING = 8


class Pathway(object):
    """
    Pathway(                       \
        pathway,                   \
        name=None                  \
        )
    COMMENT:
    Pathway(                       \
        pathway,                   \
        default_projection_matrix, \
        name=None                  \
        )
    COMMENT

    A sequence of `Nodes <Composition_Nodes>` and `Projections <Projection>` in a `Composition`, or a template
    for one that can be assigned to one or more Compositions.

    Arguments
    ---------

    pathway : list[`Node <Composition_Nodes>`, <`Projection <Projection>`,> `Node <Composition_Nodes>`...]
        specifies list of `Nodes <Composition_Nodes>` and intercolated `Projections <Projection>` to be
        created for the Pathway.

    COMMENT:
    default_projection_matrix : list, array, function, `RandomMatrix` or MATRIX_KEYWORD : default None
        specifies matrix to use for any unspecified Projections (overrides default matrix for `MappingProjection`)
        if a default projection is not otherwise specified (see `Pathway_Specification_Projections`;
        see `MappingProjection_Matrix_Specification` for details of specification)
    COMMENT

    name : str : default see `name <Pathway.name>`
        specifies the name of the Pathway (see `name <Pathway.name>` for additional information).

    Attributes
    ----------

    pathway : `Node <Composition_Nodes>`, list, tuple, or dict.
        if the Pathway is created on its own, this contains the specification provided to the **pathway** argument
        of its constructor, and can take any of the forms permitted for `Pathway specification <Pathway_Specification>`;
        if the Pathway is created by a Composition, this is a list of the `Nodes <Pathway_Nodes>` and intercolated
        `Projections <Projection>` in the Pathway (see `above <Pathway_Attribute>` for additional details).

    composition : `Composition` or None
        `Composition` to which the Pathway belongs;  if None, then Pathway is a `template <Pathway_Template>`.

    roles : list[`PathwayRole`] or None
        list of `PathwayRole(s) <PathwayRole>` assigned to the Pathway, based on the `NodeRole(s) <NodeRole>`
        assigned to its `Nodes <Composition>` in the `composition <Pathway.composition>` to which it belongs.
        Returns an empty list if belongs to a Composition but no `PathwayRoles <PathwayRole>` have been assigned,
        and None if the Pathway is a `tempalte <Pathway_Template>` (i.e., not assigned to a Composition).

    default_projection_matrix : list, array, function, `RandomMatrix` or MATRIX_KEYWORD : default None
        matrix used for any unspecified Projections (overrides default matrix for `MappingProjection`)
        if a default projection is not otherwise specified (see `Pathway_Specification_Projections`;
        see `MappingProjection_Matrix_Specification` for details of specification).  A default_projection_matrix
        is specified by including it in a tuple specification in the **pathways** argument of the Pathway's
        constructor (see `2 or 3-item tuple <Pathway_Specification_Tuple>`).

    learning_function : `LearningFunction` or None
        `LearningFunction` used by `LearningMechanism(s) <LearningMechanism>` associated with Pathway if
        it is a `learning pathway <Composition_Learning_Pathway>`.

    input : `Mechanism <Mechanism>` or None
        `INPUT` node if Pathway contains one.

    output : `Mechanism <Mechanism>` or None
        `OUTPUT` node if Pathway contains one.

    target : `Mechanism <Mechanism>` or None
        `TARGET` node if if Pathway contains one; same as `learning_components
        <Pathway.learning_components>`\\[*TARGET_MECHANISM*].

    learning_objective : `Mechanism <Mechanism>` or None
        `OBJECTIVE_MECHANISM` if Pathway contains one; same as `learning_components
        <Pathway.learning_components>`\\[*COMPATOR_MECHANISM*].

    learning_components : dict or None
        dict containing the following entries if the Pathway is a `learning Pathway <Composition_Learning_Pathway>`
        (and is assigned `PathwayRole.LEARNING` in `roles <Pathway.roles>`):

          *TARGET_MECHANISM*: `ProcessingMechanism` (assigned to `target <Pathway.target>`)
          ..
          *OBJECTIVE_MECHANISM*: `ComparatorMechanism` (assigned to `learning_objective <Pathway.learning_objective>`)
          ..
          *LEARNING_MECHANISMS*: `LearningMechanism` or list[`LearningMechanism`]
          ..
          *LEARNED_PROJECTIONS*: `Projection <Projection>` or list[`Projections <Projection>`]

        These are generated automatically and added to the `Composition` when the Pathway is assigned to it.
        Returns an empty dict if it is not a `learning Pathway <Composition_Learning_Pathway>`, and None
        if the Pathway is a `tempalte <Pathway_Template>` (i.e., not assigned to a Composition).

    name : str
        the name of the Pathway; if it is not specified in the **name** argument of the constructor, a default is
        assigned by PathwayRegistry (see `Registry_Naming` for conventions used for default and duplicate names).
        See `note <Pathway_Name_Note>` for additional information.

    """
    componentType = 'Pathway'
    componentName = componentType
    name = componentName

    def __init__(
            self,
            pathway:list,
            # default_projection_matrix=None,
            name=None,
            **kwargs
    ):

        context = kwargs.pop(CONTEXT, None)

        # Get composition arg (if specified)
        self.composition = None
        self.composition = kwargs.pop('composition',None)
        # composition arg must be a Composition
        if self.composition is not None:
            if not isinstance(self.composition, Composition):
                raise CompositionError(
                    f"'composition' arg of constructor for {self.__class__.__name__} must be a {Composition.__name__}."
                )

        # There should be no other arguments in constructor
        if kwargs:
            raise CompositionError(f"Illegal argument(s) used in constructor for {self.__class__.__name__}: "
                                   f"{', '.join(list(kwargs.keys()))}.")

        # Register and get name
        # - if called from command line, being used as a template, so don't register
        if context is None:
            # But do pass through name so that it can be used to construct the instance that will be used
            self.name = name
        else:
            # register and set name
            register_category(
                entry=self,
                base_class=Pathway,
                registry=PathwayRegistry,
                name=name
            )

        # Initialize attributes
        self.pathway = pathway
        if self.composition:
            self.learning_components = {}
            self.roles = set()
        else:
            self.learning_components = None
            self.roles = None

        # Assign default_projection_matrix attribute
        # self.default_projection_matrix = default_projection_matrix
        # Parse from tuple spec in **pathway** arg:
        self.default_projection_matrix = None
        if isinstance(self.pathway, tuple):
            for item in self.pathway:
                if is_matrix(item):
                    self.default_projection_matrix = item
        assert True

    def _assign_roles(self, composition):
        """Assign `PathwayRoles <PathwayRole>` to Pathway based `NodeRoles <NodeRole>` assigned to its `Nodes
        <Composition_Nodes>` by the **composition** to which it belongs.
        """
        assert self.composition, f'_assign_roles() cannot be called for {self.name} ' \
                                 f'because it has not been assigned to a {Composition.__name__}.'
        self.roles = set()
        for node in self.pathway:
            if not isinstance(node, (Mechanism, Composition)):
                continue
            roles = composition.get_roles_by_node(node)
            if NodeRole.ORIGIN in roles:
                self.roles.add(PathwayRole.ORIGIN)
            if NodeRole.INPUT in roles:
                self.roles.add(PathwayRole.INPUT)
            if NodeRole.SINGLETON in roles and len(self.pathway)==1:
                self.roles.add(PathwayRole.SINGLETON)
            if NodeRole.OUTPUT in roles:
                self.roles.add(PathwayRole.OUTPUT)
            if NodeRole.TERMINAL in roles:
                self.roles.add(PathwayRole.TERMINAL)
            if NodeRole.CYCLE in roles:
                self.roles.add(PathwayRole.CYCLE)
        if not [role in self.roles for role in {PathwayRole.ORIGIN, PathwayRole.TERMINAL}]:
            self.roles.add(PathwayRole.INTERNAL)
        if self.learning_components:
            self.roles.add(PathwayRole.LEARNING)

    @property
    def input(self):
        if self.composition and PathwayRole.INPUT in self.roles:
            input_node = next(n for n in self.pathway if n in self.composition.get_nodes_by_role(NodeRole.INPUT))
            if input_node:
                return input_node
            else:
                assert False, f"PROGRAM ERROR: {self.__class__.__name__} {self.name} of {self.composition.name} " \
                              f"is assigned PathwayRole.INPUT but has no INPUT node."

    @property
    def output(self):
        if self.composition and PathwayRole.OUTPUT in self.roles:
            output_node = next(n for n in self.pathway if n in self.composition.get_nodes_by_role(NodeRole.OUTPUT))
            if output_node:
                return output_node
            else:
                assert False, f"PROGRAM ERROR: {self.__class__.__name__} {self.name} of {self.composition.name} " \
                              f"is assigned PathwayRole.OUTPUT but has no OUTPUT node."

    @property
    def target(self):
        if self.composition:
            try:
                return self.learning_components[TARGET_MECHANISM]
            except:
                if PathwayRole.LEARNING not in self.roles:
                    warnings.warn(f"{self.__class__.__name__} {self.name} 'target' attribute is None "
                                  f"because it is not a learning_pathway.")
                else:
                    assert False, f"PROGRAM ERROR: {self.__class__.__name__} {self.name} of {self.composition.name} " \
                                  f"has PathwayRole.LEARNING assigned but no 'target' attribute."
                return None

    @property
    def learning_objective(self):
        if self.composition:
            try:
                return self.learning_components[OBJECTIVE_MECHANISM]
            except:
                if PathwayRole.LEARNING not in self.roles:
                    warnings.warn(f"{self.__class__.__name__} {self.name} 'learning_objective' attribute "
                                  f"is None because it is not a learning_pathway.")
                else:
                    assert False, f"PROGRAM ERROR: {self.__class__.__name__} {self.name} of {self.composition.name} " \
                                  f"has PathwayRole.LEARNING assigned but no 'learning_objective' attribute."
                return None

    @property
    def learning_function(self):
        if self.composition:
            try:
                return self.learning_components[LEARNING_FUNCTION]
            except:
                if PathwayRole.LEARNING not in self.roles:
                    warnings.warn(f"{self.__class__.__name__} {self.name} 'learning_function' attribute "
                                  f"is None because it is not a learning_pathway.")
                else:
                    assert False, f"PROGRAM ERROR: {self.__class__.__name__} {self.name} of {self.composition.name} " \
                                  f"has PathwayRole.LEARNING assigned but no 'learning_function' attribute."
                return None



import warnings

__all__ = ['PNLWarning', 'PNLInternalWarning', 'PNLUserWarning']

class PNLWarning(Warning):
    pass

class PNLInternalWarning(PNLWarning):
    pass

class PNLUserWarning(PNLWarning):
    pass


def _disable_internal_warnings():
    warnings.simplefilter("ignore", PNLInternalWarning)

def _enable_internal_warnings():
    warnings.simplefilter("default", PNLInternalWarning)


# Disable internal warnings by default
_disable_internal_warnings()



from . import agtcontrolmechanism
from . import lccontrolmechanism

from .agtcontrolmechanism import *
from .lccontrolmechanism import *

__all__ = list(agtcontrolmechanism.__all__)
__all__.extend(lccontrolmechanism.__all__)



import pandas as pd
import numpy as np
import psyneulink as pnl

def _run_ddm_with_params(
    starting_value,
    rate,
    noise,
    threshold,
    non_decision_time,
    time_step_size,
    trial_inputs,
):
    """Create a composition with DDM and run it with the given parameters."""

    # Create a simple one mechanism composition containing a DDM in integrator mode.
    decision = pnl.DDM(
        function=pnl.DriftDiffusionIntegrator(
            starting_value=starting_value,
            rate=rate,
            noise=noise,
            threshold=threshold,
            non_decision_time=non_decision_time,
            time_step_size=time_step_size,
        ),
        output_ports=[pnl.DECISION_OUTCOME, pnl.RESPONSE_TIME],
        name="DDM",
    )

    comp = pnl.Composition(pathways=decision)

    # Run the composition to generate some data to fit
    comp.run(inputs={decision: trial_inputs})
    results = comp.results

    data_to_fit = pd.DataFrame(
        np.squeeze(np.array(results)), columns=["decision", "response_time"]
    )
    data_to_fit["decision"] = data_to_fit["decision"].astype("category")

    return comp, data_to_fit


# High-level parameters the impact performance of the test
num_trials = 50
time_step_size = 0.01
num_estimates = 1000

# Let's generate an "experimental" dataset to fit. This is a parameter recovery test
# Lets make 10% of the trials have a positive stimulus drift rate, and the other 90%
# have a negative stimulus drift rate.
rng = np.random.default_rng(12345)
trial_inputs = rng.choice(
    [5.0, -5.0], size=(num_trials, 1), p=[0.10, 0.9], replace=True
)

# Make the first and last input positive for sure. This helps make sure inputs are really getting
# passed to the composition correctly during parameter fitting, and we aren't just getting a single
# trials worth of a cached input.
trial_inputs[0] = np.abs(trial_inputs[0])
trial_inputs[-1] = np.abs(trial_inputs[-1])

ddm_params = dict(
    starting_value=0.0,
    rate=0.3,
    noise=1.0,
    threshold=0.6,
    non_decision_time=0.15,
    time_step_size=time_step_size,
)

# We will generate a dataset that comprises two different conditions. Each condition will have a different
# threshold.
params_cond1 = dict(
    threshold=0.7,
)

params_cond2 = dict(
    threshold=0.3,
)

comp, data_cond1 = _run_ddm_with_params(**{**ddm_params, **params_cond1}, trial_inputs=trial_inputs)
_, data_cond2 = _run_ddm_with_params(**{**ddm_params, **params_cond2}, trial_inputs=trial_inputs)

# Combine the data from the two conditions
data_cond1['condition'] = 'cond_t=0.7'
data_cond2['condition'] = 'cond_t=0.3'
data_to_fit = pd.concat([data_cond1, data_cond2])

# Add the inputs as columns to the data temporarily so we can shuffle the data and shuffle the inputs together
data_to_fit['inputs'] = np.concatenate([trial_inputs, trial_inputs])

# Shuffle the data, seed is set for reproducibility
data_to_fit = data_to_fit.sample(frac=1, random_state=42)

# Extract the shuffled inputs
trial_inputs = data_to_fit['inputs'].to_numpy().reshape(-1, 1)
data_to_fit = data_to_fit.drop(columns='inputs')

fit_parameters = {
    ("rate", comp.nodes['DDM']): np.linspace(-0.5, 0.5, 1000),
    ("non_decision_time", comp.nodes['DDM']): np.linspace(0.0, 1.0, 1000),
    ("threshold", comp.nodes['DDM']): np.linspace(0.1, 1.0, 1000),
}

pec = pnl.ParameterEstimationComposition(
    name="pec",
    nodes=[comp],
    parameters=fit_parameters,
    depends_on={("threshold", comp.nodes['DDM']): 'condition'},
    outcome_variables=[
        comp.nodes['DDM'].output_ports[pnl.DECISION_OUTCOME],
        comp.nodes['DDM'].output_ports[pnl.RESPONSE_TIME],
    ],
    data=data_to_fit,
    optimization_function=pnl.PECOptimizationFunction(
        method="differential_evolution",
    ),
    num_estimates=num_estimates,
    initial_seed=42,
)

pec.controller.parameters.comp_execution_mode.set("LLVM")
pec.controller.function.parameters.save_values.set(True)
pec.run(inputs={comp: trial_inputs})

records = []
params = {
    'DDM.rate': ddm_params['rate'],
    'DDM.non_decision_time': ddm_params['non_decision_time'],
    'DDM.threshold[cond_t=0.3]': 0.3,
    'DDM.threshold[cond_t=0.7]': 0.7
}
for i, (name, recovered_param) in enumerate(pec.optimized_parameter_values.items()):
    percent_error = 100.0 * (abs(params[name] - recovered_param) / params[name])
    records.append((name, params[name], recovered_param, percent_error))
df = pd.DataFrame(records, columns=['Parameter', 'Value', 'Recovered Value', 'Percent Error'])
print(df)

